;BootVBR.asm - Volume Boot Record for Thunder, v0.1 Alpha
;Created by Timothy Williams on 041417-1722

bits 16
org 0x7C00

;This area is the BPB or Bios Parameter Block, it holds information about the filesystem
;When the VBR is loaded it includes the BPB, which must be jumped over, the first three bytes do this.
;Once in our main code the third byte can easily be used as a flag byte or counter, but I don't use it
start:
	jmp short Code_Start					;Jump over BPB,This instuction should be only 2 bytes long
	nop										;Used as padding
BPB_OEM_ID 				db "MSDOS5.0"		;Must be 8 bytes long
BPB_BYTES_PER_SECTOR	dw 0x0200			;Bytes Per sector, Usually 512
BPB_SECTORS_PER_CLUSTER db 0x00				;Sectors Per Cluster
BPB_RESERVED_SECTORS	dw 0x0000			;Number of Reserved Sectors, Including Boot Record, Logical start of FAT
BPB_TOTAL_FATS			db 0x02				;Number of FAT tables, almost always 2
BPB_DIRECTORY_ENTRIES   dw 0x0000			;Number of Directory Entries
BPB_TOTAL_SECTORS		dw 0x0000			;Total sectors in the logical volume, if 0 it means there is over 65535
BPB_MEDIA_TYPE			db 0x00				;Media Descriptor Type
BPB_SECTORS_PER_FAT		dw 0x0000 			;Sectors per FAT,FAT12/FAT16 only
BPB_SECTORS_PER_TRACK   dw 0x0000 			;Sectors per track
BPB_TOTAL_HEADS			dw 0x0000			;Number of heads or sides on the storage media
BPB_HIDDEN_SECTORS		dd 0x00000000		;Number of hidden sectors, LBA of beginning of partition
BPB_LARGE_TOTAL_SECTORS dd 0x00000000		;Large amount of sectors on media, set if over 65535

;Extended Bios Parameter Block(EBPB), used in FAT32
EBPB_SECTORS_PER_FAT	dd 0x00000000 		;Sectors per FAT
EBPB_FLAGS				dw 0x0000			;Flags
EBPB_FAT_VER			dw 0x0000			;FAT Version number
EBPB_ROOT_DIR_CLUSTER   dd 0x00000002		;The cluster number of the root directory, usually 2
EBPB_FSINFO_LBA    		dw 0x0000			;The Logical LBA of the FSInfo structure
EBPB_BACKUP_VBR_LBA     dw 0x0000			;The Logical LBA of the backup boot sector
EBPB_RESERVED times 12 	db 0x00				;RESERVED, 12 Bytes
EBPB_DISK_NUM			db 0x00				;Drive number, value is arbitrary
EBPB_NT_FLAGS			db 0x00				;Flags in Windows NT, Reserved
EBPB_SIGNATURE	   		db 0x29				;Signature, must be 0x28 or 0x29
EBPB_VOLUME_ID    		dd 0x00000000		;VolumeID 'Serial' number, used for tracking volumes between computers
EBPB_VOLUME_LABEL 		db "NO NAME    "	;Volume label string, 11 Bytes
EBPB_SYS_ID 	   		db "FAT32   "		;System identifier string, 8 Bytes

;Code actually starts here
;Check to see if a MBR is present, otherwise continue
Code_Start:
	jnc No_MBR								;Skip following code if MBR is not present
	clc										;Clear Carry Flag
	mov si,bp								;Point SI to Partition Entry in MBR
	mov dl,BYTE [si]						;Get Status(I.E. Disk Number)
	mov eax,DWORD [si+0x08]					;Get First_LBA
	mov ebx,DWORD [si+0x0C]					;Get Total_Sectors

	mov DWORD [BPB_HIDDEN_SECTORS],eax		;BPB_HIDDEN_SECTORS = Start of partition

	mov	ax,WORD [BPB_TOTAL_SECTORS]			;Check BPB_TOTAL_SECTORS
	or ax,ax								;BPB_TOTAL_SECTORS != 0?
	jnz No_MBR								;Leave BPB_LARGE_TOTAL_SECTORS alone
	mov DWORD [BPB_LARGE_TOTAL_SECTORS],ebx	;BPB_LARGE_TOTAL_SECTORS = Total_Sectors

;Code Starts here if no MBR is found. Setup segments,stack,ect
No_MBR:
	cli										;Clear interrupts
	xor ax,ax								;Clear AX
	mov ss,ax								;Clear SS
	mov sp,0x7C00							;Reset stack @ 0000:7C00
	push ax									;Push two empty
	push ax									; Words
	pop ds									;Clear DS
	pop es 									;Clear ES
	sti										;Re-enable interrupts

	mov BYTE [EBPB_DISK_NUM],dl				;Ensure Disk Number is correct

;In the MBR code I check to see if the INT13h extensions are supported, I do that here as well but if they are not
;then I don't use the standard functions, instead it causes an error
	mov ah,0x41								;INT 13h AH=41h: Check Extensions Present
	mov bx,0x55AA							;Must be 55AA
	mov dl,BYTE [EBPB_DISK_NUM]				;Get Disk Number
	int 0x13								;Check for Extensions
	jc No_Ext								;Nope, Error
	cmp bx,WORD 0xAA55						;BX = AA55?
	jne No_Ext								;Nope, Error

;Now we check the filesystem for defects
;First of all is BPB_SECTORS_PER_FAT non-zero, it is a FAT12/16 field
	xor eax,eax								;Clear EAX
	cmp WORD [BPB_SECTORS_PER_FAT],ax		;Non-Zero? Probably FAT12/16 or invalid
	jnz Disk_Error							;Error

;Next are the Total sector fields non-zero
	cmp WORD [BPB_TOTAL_SECTORS],ax			;BPB_TOTAL_SECTORS = 0?
	jnz Check_Sig							;No,skip next part
	cmp DWORD [BPB_LARGE_TOTAL_SECTORS],eax ;If BPB_TOTAL_SECTORS and BPB_LARGE_TOTAL_SECTORS are zero
	jz Disk_Error							;Something is wrong

;The disk signature should be either 0x28 or 0x29, if it isn't something isn't right
Check_Sig:
	mov al,BYTE [EBPB_SIGNATURE]			;AL = EBPB_SIGNATURE, Should be either 0x28 or 0x29
	mov ah,0x28								;AH = 0x28
	cmp ah,al								;EBPB_SIGNATURE = 0x28?
	je Load_Stage2							;Yes? Ok
	inc ah									;AH = 0x29
	cmp ah,al								;EBPB_SIGNATURE = 0x29?
	je Load_Stage2							;Nope, fall into error handler

;The error handler is very similar to the one in the MBR
;There are only 3 error messages
Disk_Error:
	lea si,[DISK_ERROR]						;"Disk error!"
	jmp Error_Handler

Stage2_Missing:
	lea si,[STAGE2_MISSING]					;"Unable to Find Stage2!"
Error_Handler:
	call PrintS								;Print Error Message
	lea si,[PRESS_KEY]						;*NEWLINE*"Press any key to reboot...."
	call PrintS								;Print string
											;Await keystroke and warm reboot
	cbw										;Copy AL into AH, since 0x00 is the terminating byte, AX = 0000
	int 0x16								;INT16h AH=00h: Read keyboard input
	mov di,0x0472							;ES:DI => 0000:0472
	mov [es:di],WORD 0x1234					;Simulates pressing Ctrl-Alt-Del
	push 0xFFFF								;Segment:FFFF
	push 0x0000								;Offset :0000
	retf									;Far return to FFFF:0000, warm rebooting the computer

;PrintS - Print Stirng
;In : DS:SI => Null terminated string
PrintS:
	lodsb									;SI=>AL, Increment SI
	or al,al								;Null Byte?
	jz PrintS_Return						;Yes,Return
	mov ah,0x0E								;AH=0Eh Teletype output
	mov bx,0x0007							;BH = Page BL=Color
	int 0x10								;Print character
	jmp PrintS								;Next....
PrintS_Return:
	ret

Load_Stage2:
	mov eax,DWORD [BPB_HIDDEN_SECTORS]		;Get start of partition
	add eax,0x0C							;Stage2_LBA = BPB_HIDDEN_SECTORS+0x0C

	xor edx,edx								;EDX:EAX=Stage2_LBA
	mov bx,0x8000							;Load stage2 @ 0000:8000
	mov cx,0x02								;Read two sectors
	call Read_Disk							;Read disk

;Now that Stage 2 is loaded we need to verify if it is valid
;The signature is the last two bytes of the second sector, offset 0x83FE
;Because Windows NT will often write bootloader code @ LBA 0x0C, even if the bootloader isn't installed
; the signature has been changed to 0x29A so that VBR doesn't attempt to load Windows NT code
	mov si,0x83FE							;SI=>Signature
	cmp [si],WORD 0x29A					;Is signature valid
	jne Stage2_Missing						;Nope? Error

	push cs									;Push CS
	push 0x8000								;CS:IP=>0000:8000
	retf									;Jump to stage2 @ 0000:8000

;Here's a slightly modified version of the read disk function from the MBR
;It works about the same except you call/ret instead of jumping in and out of it
;I derped out in the previous version of this code and used DI instead of BX to pass the offset
; and I had to work in some debugging code in order to figure out what was wrong
Read_Disk:
	mov DWORD [DAP_LBA_MSD],edx				;EDX = MSD of LBA
	mov DWORD [DAP_LBA_LSD],eax				;EAX = LSD of LBA
	mov WORD [DAP_Segment],es				;ES = Segment
	mov WORD [DAP_Offset],bx				;BX = Offset
	mov WORD [DAP_Sectors],cx				;CX = Sectors
	mov cx,0x05								;Try to read disk 5 times
Read:
	mov ah,0x42								;INT 13h AH=42h: Extended Read Sectors From Drive
	mov dl,BYTE [EBPB_DISK_NUM]				;DL = Drive Number
	lea si,[DAP_Size]						;DS:SI = Offset of DAP
	int 0x13								;Read disk
	jnc Read_Return							;Return

	or cx,cx								;count = 0?
	jz Disk_Error							;Error
	dec cx									;count--

	xor ah,ah								;INT 13h AH=00h: Reset Disk Drive
	int 0x13								;Reset Disk
	jc Disk_Error							;Something's wrong
	jmp Read								;Try again
Read_Return:
	ret										;Return
;End Of Code

DISK_ERROR         db "Disk error", 0
STAGE2_MISSING     db "Second Stage not found! System halted.", 0
NO_EXT             db "Interrupt 13h extensions are not supported.", 0
PRESS_KEY          db "Press any key to restart...", 0

DAP_Size			db 0x10					;Size of packet, always 16 BYTES
DAP_Reserved		db 0x00					;Reserved
DAP_Sectors			dw 0x0000				;Number of sectors to read
DAP_Offset			dw 0x0000				;Offset to read data to
DAP_Segment 		dw 0x0000				;Segment to read data to
DAP_LBA_LSD			dd 0x00000000			;QWORD with LBA of where to start reading from
DAP_LBA_MSD			dd 0x00000000

times 510-($-$$) 	db 0
dw 0xAA55
